# Copyright 2021 Silicon Laboratories Inc. www.silabs.com
#
# SPDX-License-Identifier: Zlib
#
# The licensor of this software is Silicon Laboratories Inc.
#
# This software is provided 'as-is', without any express or implied
# warranty. In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

import time
from collections import deque
from threading import Lock, Event

from bgapi.connector import Connector

def sleep(seconds):
    if seconds > 0:
        time.sleep(seconds)


class TestConnectorException(Exception): pass


class TestConnector(Connector):

    ACTIONS = ["<", ">", "@"]

    def __init__(self):
        self.read_timeout = 0.0

        self.steps = deque()
        self.get_step_lock = Lock()
        self.new_step = Event()
        self.got_step = Event()

        super(TestConnector, self).__init__()

    def open(self):
        return

    def close(self):
        return

    def write(self, data):
        while data:
            self._process_wait()

            write_step = self._get_step(">")
            if not write_step:
                raise TestConnectorException("Unexpected write %r" % data)

            verify_data = write_step[1]
            # If not enough data received, put rest of the verify_data back to the queue
            if len(data) < len(verify_data):
                self.steps.appendleft((">", verify_data[len(data):]))

            for i, (byte, verify_byte) in enumerate(zip(data, verify_data), start=1):
                if byte != verify_byte:
                    raise TestConnectorException("Invalid write: %r != %r" % (data[:i], verify_data[:i]))

            # If too much data received, reiterate with rest of the data
            data = data[len(verify_data):]

        self._process_wait() # Do for next step also

    def read(self, size=1):
        start_time = time.time()
        retval = b""

        timeout_left = lambda: self.read_timeout - (time.time() - start_time)

        while size:
            wait = self._process_wait()
            if wait:
                if wait > timeout_left():
                    sleep(timeout_left())
                    break
                sleep(wait)
                continue

            # Wait for read step           
            read_step = self._get_step(action="<", timeout=timeout_left())
            if not read_step: break

            read_data = read_step[1]
            # If not enough data requested, put rest of the read_data back to the queue
            if size < len(read_data):
                self.steps.appendleft(("<", read_data[size:]))
                read_data = read_data[:size]

            retval += read_data
            size -= len(read_data)

        self._process_wait() # Do for next step also

        return retval

    def set_read_timeout(self, timeout):
        self.read_timeout = timeout

    def set_write_timeout(self, timeout):
        pass

    def _process_wait(self):
        time_to_wait = 0.0

        wait_step = self._get_step("@")
        if wait_step:
            wait_arg = wait_step[1]
            try:
                (wait_time, started) = wait_arg
                # Wait already started
                waited = time.time() - started
                if waited < wait_time:
                    # Wait some more
                    self.steps.appendleft(("@", (wait_time - waited, time.time())))
                    time_to_wait = wait_time - waited
                else:
                    # Waited enough
                    pass
            except TypeError:
                # Start wait
                self.steps.appendleft(("@", (wait_arg, time.time())))
                time_to_wait = wait_arg

        return time_to_wait

    def _get_step(self, action=None, timeout=0.0):
        start_time = time.time()
        timeout_left = lambda: timeout - (time.time() - start_time)

        while True:
            try:
                self.get_step_lock.acquire()
                next_step = self.steps.popleft()
                next_action = next_step[0]

                if action == None or next_action == action:
                    self.get_step_lock.release()
                    self.got_step.set()
                    return next_step

                # Put the step back to the queue
                self.steps.appendleft(next_step)
                self.get_step_lock.release()

                # Wait for someone to get the action out of the queue
                self.got_step.clear()
                self.got_step.wait(timeout_left())
                self.got_step.clear()

            except IndexError:
                self.get_step_lock.release()
                # Wait for new step to arrive
                self.new_step.clear()
                self.new_step.wait(timeout_left())
                self.new_step.clear()

            if timeout != None and timeout_left() <= 0:
                break

    def add_step(self, action, arg):
        if action not in self.ACTIONS:
            raise ValueError("Invalid action '%s'" % action)
        self.steps.append((action, arg))
        self.new_step.set()

    def add_step_str(self, *steps):
        for step in steps:
            (action, arg) = step.split(":", 1)

            if action == "@":
                arg = float(arg)

            self.add_step(action, arg)



# Copyright 2021 Silicon Laboratories Inc. www.silabs.com
#
# SPDX-License-Identifier: Zlib
#
# The licensor of this software is Silicon Laboratories Inc.
#
# This software is provided 'as-is', without any express or implied
# warranty. In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

import serial
from .connector import *


class SerialConnector(Connector):

    def __init__(self, port, baudrate=115200, rtscts=False):
        self.s = serial.Serial(baudrate=baudrate, rtscts=rtscts)
        self.s.port = port # port given separately to prevent automatic opening

        super(SerialConnector, self).__init__()

    def open(self):
        try:
            self.s.open()
        except serial.SerialException as e:
            raise ConnectorException(e)

    def close(self):
        # Temporarily disable RTS/CTS because in some cases, closing the serial
        # port takes 30 seconds to complete if handshaking is enabled, at least
        # with Linux and FT232 USB to serial adapters.
        rtscts = self.s.rtscts
        self.s.rtscts = False
        try:
            self.s.close()
        finally:
            self.s.rtscts = rtscts

    def write(self, data):
        try:
            self.s.write(data)
        except ValueError as e:
            raise ConnectorException(e)
        except serial.SerialTimeoutException as e:
            raise ConnectorTimeoutException(e)

    def read(self, size=1):
        try:
            return self.s.read(size)
        except ValueError as e:
            raise ConnectorException(e)

    def set_read_timeout(self, timeout):
        self.s.timeout = timeout

    def set_write_timeout(self, timeout):
        self.s.writeTimeout = timeout



# Copyright 2021 Silicon Laboratories Inc. www.silabs.com
#
# SPDX-License-Identifier: Zlib
#
# The licensor of this software is Silicon Laboratories Inc.
#
# This software is provided 'as-is', without any express or implied
# warranty. In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

import struct
import binascii

import bgapi.serdeser

def str2bin(string):
    return binascii.unhexlify(string)


def bin2str(binary):
    return binascii.hexlify(binary)


def hexstring(length):
    return b"".join([b"%02x" % (i % 256) for i in range(length)])


make_header = bgapi.serdeser.make_header


def make_packet(msg_type, device_id, class_id, command_id, payload):
    payload = str2bin(payload)

    header = make_header(msg_type, device_id, class_id, command_id, len(payload))
    return (header, payload)



# Copyright 2021 Silicon Laboratories Inc. www.silabs.com
#
# SPDX-License-Identifier: Zlib
#
# The licensor of this software is Silicon Laboratories Inc.
#
# This software is provided 'as-is', without any express or implied
# warranty. In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

"""
Takes care of serializing and deserializing of api commands and events
"""

import binascii
import struct
import logging
import operator

from .apiparser import toInt


logger = logging.getLogger(__package__)

MSG_COMMAND = 0
MSG_EVENT = 1

HEADER_LENGTH = 4


def to_bytes(s):
    if isinstance(s, bytes):
        return s
    else:
        return s.encode("latin-1")


def make_header(msg_type, device_id, class_id, command_id, payload_len):
    header = [
            (msg_type << 7) | (device_id << 3) | (payload_len >> 8),
            payload_len & 0xff,
            class_id,
            command_id,
    ]
    return struct.pack("<%dB" % HEADER_LENGTH, *header)


class Serializer(object):

    def __init__(self, apis):
        self.apis = {}
        self.device_ids = {}
        for api in apis:
            self.apis[api.device_id] = api
            self.device_ids[api.device_name] = api.device_id

    def _convertEnumDefine(self, apiparam, val, enums, defines):
        """
        Converts textual definitions into their numeric versions
        """
        if not apiparam.validator_id:
            return toInt(val)
        if not isinstance(val, basestring) or  len(val) == 0 or val[0] in "0123456789":
            return toInt(val)

        if apiparam.validator_type == "enum":
            val = enums[val].value
        elif apiparam.validator_type == "define":
            vals = val.split("|")
            val = 0
            for v in vals:
                val |= defines[v].value
        return toInt(val)

    def _parseParameters(self, args, apiparams, enums, defines):
        """ Returns tuple (pack format list, argument list)  """
        if len(args) != len(apiparams):
            raise TypeError("Expected %d arguments, %d given (%r)" % (len(apiparams), len(args), args))

        packers = {
            # {format : (pack code, python value to pack) }
            "int8" :
                lambda apiparam, x: ("b", self._convertEnumDefine(apiparam, x, enums, defines)),
            "uint8" :
                lambda apiparam, x: ("B", self._convertEnumDefine(apiparam, x, enums, defines)),
            "int16" :
                lambda apiparam, x: ("h", self._convertEnumDefine(apiparam, x, enums, defines)),
            "uint16" :
                lambda apiparam, x: ("H", self._convertEnumDefine(apiparam, x, enums, defines)),
            "int32" :
                lambda apiparam, x: ("i", self._convertEnumDefine(apiparam, x, enums, defines)),
            "uint32" :
                lambda apiparam, x: ("I", self._convertEnumDefine(apiparam, x, enums, defines)),
            "int64" :
                lambda apiparam, x: ("q", self._convertEnumDefine(apiparam, x, enums, defines)),
            "uint64" :
                lambda apiparam, x: ("Q", self._convertEnumDefine(apiparam, x, enums, defines)),
            "uint8array" :
                lambda apiparam, x: ("%ds" % (len(to_bytes(x)) + 1), struct.pack("<B", len(x)) + to_bytes(x)),
            "uint16array" :
                lambda apiparam, x: ("%ds" % (len(to_bytes(x)) + 2), struct.pack("<H", len(x)) + to_bytes(x)),
            "bd_addr" :
                lambda apiparam, x: ("6s", binascii.unhexlify(to_bytes(x).replace(b":", b""))[::-1]), #Last part is to reverse byte order
            "hw_addr" :
                lambda apiparam, x: ("6s", binascii.unhexlify(to_bytes(x).replace(b":", b""))),
            "ipv4" :
                lambda apiparam, x: ("4s", b"".join(map(struct.Struct(">B").pack, map(int, x.split("."))))),
            "uuid_128" :
                lambda apiparam, x: ("16s", to_bytes(x)),
            "aes_key_128" :
                lambda apiparam, x: ("16s", to_bytes(x)),
            "sl_bt_uuid_64_t" :
                lambda apiparam, x: ("8s", to_bytes(x)),

        }

        a = tuple(zip(*[ packers[apiparam.format](apiparam, val) for apiparam, val in zip(apiparams, args) ]))
        if not a:
            return ([], [])

        return a

    def command(self, device_id, class_name, command_name, args):
        if isinstance(device_id, str):
            device_id = self.device_ids[device_id]

        apiclass = self.apis[device_id][class_name]
        apicommand = apiclass.commands[command_name]

        # parameters into binary
        pack_format, params = self._parseParameters(args, apicommand.params, apiclass.enums, apiclass.defines)
        pack_format = "".join(pack_format)
        param_data_len = struct.calcsize("<%s" % pack_format)

        header = make_header(MSG_COMMAND, device_id, apiclass.index,
                             apicommand.index, param_data_len)
        payload = struct.pack("<%s" % pack_format, *params)

        return header + payload


class Deserializer(object):

    def parseHeader(self, header):
        """
        Returns tuple with:
            - type: COMMAND=0 or EVENT=1
            - device id: BLE = 0
            - parameter length
            - class id
            - command id
        """
        return (
            (operator.getitem(header, 0) & 0x80) >> 7,
            (operator.getitem(header, 0) & 0x78) >> 3,
            (operator.getitem(header, 0) & 0x7) << 8 | operator.getitem(header, 1),
            operator.getitem(header, 2),
            operator.getitem(header, 3)
            )

    def __init__(self, apis):
        self.apis = {}
        for api in apis:
            self.apis[api.device_id] = api

    def _convertEnumDefine(self, apiparam, val, enums, defines):
        if not apiparam.validator_id:
            return val
        if apiparam.validator_type == "enum":
            return enums[apiparam.validator_id][val].name
        elif apiparam.validator_type == "define":
            ret = []
            for flag in apiparam.defines[apiparam.validator_id]:
                if flag.value & val: ret.append(flag.name)
            if len(ret) != 0:
                return "|".join(ret)
        return val

    def parse(self, header, payload, fromHost, enumdefines=True):
        """
            Returns a tuple consisting of: (apicommand, header field tuple, param list)
        """
        hfields = self.parseHeader(header)

        cmdevt, device_id, parameter_len, classIndex, cmdEvtIndex = hfields
        try:
            api = self.apis[device_id]
        except KeyError:
            raise Exception("No API definition for device ID %d" % device_id)
        apidoc_name = api.name

        if cmdevt == MSG_COMMAND:
            cmdevt = api[classIndex].commands[cmdEvtIndex]
            if fromHost:
                apiparams = cmdevt.params
                apidoc_name += "_cmd"
            else:
                apiparams = cmdevt.returns
                apidoc_name += "_rsp"
        else:
            cmdevt = api[classIndex].events[cmdEvtIndex]
            apiparams = cmdevt.params
            apidoc_name += "_evt"

        apidoc_name += "_%s_%s" % (cmdevt.api_class.name, cmdevt.name)

        unpackers = {
            "int8": lambda x: "b",
            "uint8": lambda x: "B",
            "int16": lambda x: "h",
            "uint16": lambda x: "H",
            "int32": lambda x: "i",
            "uint32": lambda x: "I",
            "int64": lambda x: "q",
            "uint64": lambda x: "Q",
            "uint8array": lambda x: "x%ds" % struct.unpack("<B", x[:1]),
            "uint16array": lambda x: "xx%ds" % struct.unpack("<H", x[:2]),
            "bd_addr": lambda x: "6s",
            "hw_addr": lambda x: "6s",
            "ipv4": lambda x: "4s",
            "uuid_128": lambda x: "16s",
            "aes_key_128": lambda x: "16s",
            "sl_bt_uuid_64_t": lambda x: "8s",
        }

        pos = 0
        pack_format = ""
        for (i, apiparam) in enumerate(apiparams):
            if pos >= len(payload):
                # Message has less params than in the API definition
                missing_params = [param.name for param in apiparams]
                logger.warning(
                        "Received message '%s' with parameter(s) %s missing.",
                        apidoc_name,
                        ", ".join(["'%s'" % (p.name) for p in apiparams[i:]]))
                break
            # Give first 2 bytes of the field for array unpackers
            fmt = unpackers[apiparam.format](payload[pos:pos+2])
            pack_format += fmt
            pos += struct.calcsize(fmt)

        if len(payload) > pos:
            # Message has more params than in the API definition, ignore extra
            logger.warning("Received message '%s' with %d byte(s) extra payload.",
                    apidoc_name, len(payload) - pos)
            payload = payload[:pos]

        vals = struct.unpack("<%s" % pack_format, payload)

        params = list(vals)

        # If message had less params than defined in the API, put None in place
        missing_params = len(apiparams) - len(params)
        params += [None] * missing_params

        converters = {
            "int8": lambda x: x,
            "uint8": lambda x: x,
            "int16": lambda x: x,
            "uint16": lambda x: x,
            "int32": lambda x: x,
            "uint32": lambda x: x,
            "int64": lambda x: x,
            "uint64": lambda x: x,
            "uint8array": lambda x: x,
            "uint16array": lambda x: x,
            "bd_addr": lambda x: ":".join(["{:02x}".format(b) for b in iter(reversed(x))]),
            "hw_addr": lambda x: ":".join(["{:02x}".format(b) for b in iter(x)]),
            "ipv4": lambda x: ".".join(map(str, iter(x))),
            "uuid_128": lambda x: x,
            "aes_key_128": lambda x: x,
            "sl_bt_uuid_64_t": lambda x: x,
        }

        params = [converters[apiparam.format](param) for (param, apiparam) in zip(params, apiparams)]

        if enumdefines:
            enums = api[classIndex].enums
            defines = api[classIndex].defines
            params = [self._convertEnumDefine(apiparams[i], params[i], enums , defines) for i in range(len(apiparams))]

        return (cmdevt, hfields, params)



# Copyright 2021 Silicon Laboratories Inc. www.silabs.com
#
# SPDX-License-Identifier: Zlib
#
# The licensor of this software is Silicon Laboratories Inc.
#
# This software is provided 'as-is', without any express or implied
# warranty. In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

import os
import struct
import unittest

from bgapi.apiparser import ParsedApi
from bgapi.serdeser import Serializer, Deserializer, MSG_COMMAND, MSG_EVENT, HEADER_LENGTH

from .helper import *

class TestTypes(unittest.TestCase):

    def setUp(self):
        self.api = ParsedApi(os.path.join(os.path.dirname(__file__), "test.xml"))
        self.ser = Serializer([self.api])
        self.deser = Deserializer([self.api])

    def verify(self, msg, string):
        self.assertEqual(bin2str(msg), string)

    def verify_payload(self, msg, string):
        self.assertEqual(bin2str(msg[HEADER_LENGTH:]), string)

    def ser_command(self, class_name, command_name, *params):
        return self.ser.command(self.api.device_id, class_name, command_name, params)

    def deser_response(self, class_name, command_name, payload):
        payload = str2bin(payload)
        class_id = self.api[class_name].index
        command_id = self.api[class_name].commands[command_name].index
        header = make_header(MSG_COMMAND, self.api.device_id, class_id, command_id, len(payload))
        return self.deser.parse(header, payload, fromHost=False)

    def ser_type_test(self, type, value, result):
        cmd = self.ser_command("types", type, value)
        self.verify_payload(cmd, result)

    def deser_type_test(self, type, value, binary):
        (cmdevt, hfields, params) = self.deser_response("types", type, binary)
        self.assertEqual(params[0], value)

    def type_test(self, type, value, binary):
        self.ser_type_test(type, value, binary)
        self.deser_type_test(type, value, binary)

    # Actual tests start from here

    def test_int8(self):
        self.type_test("int8", -1, b"ff")
        self.type_test("int8", 127, b"7f")
        self.type_test("int8", -128, b"80")

    def test_uint8(self):
        self.type_test("uint8", 0x12, b"12")

    def test_int16(self):
        self.type_test("int16", -1, b"ffff")
        self.type_test("int16", 32767, b"ff7f")
        self.type_test("int16", -32768, b"0080")

    def test_uint16(self):
        self.type_test("uint16", 0x1234, b"3412")

    def test_int32(self):
        self.type_test("int32", -2147483648, b"00000080")
        self.type_test("int32", -305419896, b"88a9cbed")
        self.type_test("int32", -1, b"ffffffff")
        self.type_test("int32", 0, b"00000000")
        self.type_test("int32", 1, b"01000000")
        self.type_test("int32", 305419896, b"78563412")
        self.type_test("int32", 2147483647, b"ffffff7f")

    def test_uint32(self):
        self.type_test("uint32", 0x12345678, b"78563412")

    def test_int64(self):
        self.type_test("int64", -9223372036854775808, b"0000000000000080")
        self.type_test("int64", -1, b"ffffffffffffffff")
        self.type_test("int64", 0, b"0000000000000000")
        self.type_test("int64", 1, b"0100000000000000")
        self.type_test("int64", 81985529216486895, b"efcdab8967452301")
        self.type_test("int64", 9223372036854775807, b"ffffffffffffff7f")

    def test_uint64(self):
        self.type_test("uint64", 0x0123456789abcdef, b"efcdab8967452301")

    def test_uint8array(self):
        self.type_test("uint8array", str2bin(hexstring(0)), b"00" + hexstring(0))
        self.type_test("uint8array", str2bin(hexstring(5)), b"05" + hexstring(5))
        self.type_test("uint8array", str2bin(hexstring(255)), b"ff" + hexstring(255))

    def test_uint16array(self):
        self.type_test("uint16array", str2bin(hexstring(0)), b"0000" + hexstring(0))
        self.type_test("uint16array", str2bin(hexstring(5)), b"0500" + hexstring(5))
        self.type_test("uint16array", str2bin(hexstring(255)), b"ff00" + hexstring(255))
        self.type_test("uint16array", str2bin(hexstring(256)), b"0001" + hexstring(256))
        # Test maximum size of an array allowed by BGAPI packet format
        self.type_test("uint16array", str2bin(hexstring(2045)), b"fd07" + hexstring(2045))

    def test_bd_addr(self):
        self.type_test("bd_addr", "12:34:56:78:90:ab", b"ab9078563412")

    def test_hw_addr(self):
        self.type_test("hw_addr", "12:34:56:78:90:ab", b"1234567890ab")

    def test_ipv4(self):
        self.type_test("ipv4", "18.52.86.120", b"12345678")

    def test_uuid_128(self):
        self.type_test("uuid_128", b"Pepe the Frog 88", b"50657065207468652046726f67203838")

    def test_aes_key_128(self):
        self.type_test("aes_key_128", b"\x00\x01\x02\x03\x04\x05\x06\x07\x00\x01\x02\x03\x04\x05\x06\x07", b"00010203040506070001020304050607")

    def test_uuid_64(self):
        self.type_test("sl_bt_uuid_64_t", b"Harambe!", b"486172616d626521")


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Silicon Laboratories Inc. www.silabs.com
#
# SPDX-License-Identifier: Zlib
#
# The licensor of this software is Silicon Laboratories Inc.
#
# This software is provided 'as-is', without any express or implied
# warranty. In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

class Connector(object):
    """
    Connector class abstracts data transmission between arbitrary end-points, like
    serial-ports or TCP/IP-sockets.
    """

    def open(self):
        """
        Open connector. Raises ConnectorException on failure.
        """
        raise NotImplementedError()

    def close(self):
        """
        Close connector. Shouldn't fail.
        """
        raise NotImplementedError()

    def write(self, data):
        """
        Write data to connector.
        """
        raise NotImplementedError()

    def read(self, size=1):
        """
        Read data from connector. Read blocks until received size number of bytes.
        """
        raise NotImplementedError()

    def set_read_timeout(self, timeout):
        """
        Sets the timeout in seconds for read operations.
        """
        raise NotImplementedError()

    def set_write_timeout(self, timeout):
        """
        Sets the timeout in seconds for write operations.
        """
        raise NotImplementedError()


class ConnectorException(Exception):
    """ Base class for connector related exceptions. """

class ConnectorTimeoutException(ConnectorException):
    """ Timeouts give this exception. """



# Copyright 2021 Silicon Laboratories Inc. www.silabs.com
#
# SPDX-License-Identifier: Zlib
#
# The licensor of this software is Silicon Laboratories Inc.
#
# This software is provided 'as-is', without any express or implied
# warranty. In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

import re


def camelcase(text):
    return "".join([x.capitalize() for x in text.split("_")])


def strip_html(description, singleLine = False):
    return re.sub("<.*?>", "", description).strip()


def api_cmd_to_ascii(cmdnode):
    description = strip_html(cmdnode.description)
    if not description:
        description = "Send command %s." % cmdnode.name

    text = [description]
    text.append("\nArguments:")
    if not cmdnode.params:
        text.append(" (none)")
    for param in cmdnode.params:
        text.append("\t - %s: %s"%(param.name, strip_html(param.description)))
    if hasattr(cmdnode, "returns") and cmdnode.returns:
        text.append("\nReturn values:")
        for ret in cmdnode.returns:
            if ret.description:
                text.append(" - %s: %s"%(ret.name, strip_html((ret.description))))
            else:
                text.append(" - %s"%ret.name) 
    return "\n".join(text)


class ApiHelper(object):
    def __init__(self, api):
        self.api = api
        self.classes = self.api.keys()

    def split_user_input(self, text):
        """
        Returns ( class, command, [param1, param2, ...] )
        """
        apiclass, rest = text.split("_",1)
        if " " in rest:
            command, rest = rest.split(" ", 1)
        else:
            command = rest
            rest = ""
        params = rest.split(" ")
        return (apiclass, command, params)

    def complete_enum_define(self, apiclass, apiparam, val):
        """
        Takes the param to be suggested and current val
        returns a sorted list of suggestions
        """
        if apiparam.validator_id == None: return []
        if apiparam.validator_type == "enum":
            possible_vals = apiclass.enums[apiparam.validator_id].keys()
        elif apiparam.validator_type == "define":
            if "|" in val: val = val.split("|")[-1]
            possible_vals = apiclass.defines[apiparam.validator_id].keys()
        else:
            possible_vals = []
        return sorted(filter(lambda x: x.startswith(val), possible_vals))

    def is_current_param_complete(self, userinput):
        return userinput.endswith(" ")

    def is_command_complete(self, userinput):
        return " " in userinput

    def is_class_complete(self, userinput):
        return "_" in userinput

    def completion_suggestions(self, text):
        class_name, command_name, params =  self.split_user_input(text)
        if not self.is_class_complete(text):
            return sorted(filter(lambda x: x.startswith(apiclass), self.api.keys()))
        if class_name not in self.api:
            return []
        if not self.is_command_complete(text):
            command_suggestions = sorted(filter(lambda x: x.startswith(command_name), self.api[class_name].command_names))
            return ["%s_%s"%(class_name, x) for x in command_suggestions]
        if command_name not in self.api[class_name].commands:
            return []

        if params[0] == "":
            cur_param = 0
        else:
            cur_param = len(params)-1

        if cur_param >= len(param_count):
            return []

        apiclass = self.api[apiclass]
        apiparam = apiclass.commands[command].prams[cur_param]
        return complete_enum_define(self, apiclass, apiparam, params[cur_param])



# Copyright 2021 Silicon Laboratories Inc. www.silabs.com
#
# SPDX-License-Identifier: Zlib
#
# The licensor of this software is Silicon Laboratories Inc.
#
# This software is provided 'as-is', without any express or implied
# warranty. In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

import socket
import warnings

from .connector import *


class SocketConnector(Connector):
    """Connector for TCP/IP or Unix domain socket connections

    The address format depends on the selected address family. See the
    socket module documentation in the Python standard library for more
    details about supported address families. The family parameter supports
    constants from the socket module, e.g. socket.AF_INET, or the family
    name may be passed as a string, which is converted to the correct type.

    For TCP/IP connections, the address is a tuple (host, port).
    Example:
        >>> SocketConnector(('127.0.0.1', 1234))
        >>> SocketConnector(('127.0.0.1', 1234), family='inet')
        >>> SocketConnector(('::1', 1234), family='inet6')

    For Unix domain socket connections, the address is a local filesystem
    path.
    Example:
        >>> SocketConnector('/path/to/bgapi.sock', family='unix')

    NOTE: The port parameter is deprecated and should not be used anymore!
    """

    def __init__(self, address, port=None, family=socket.AF_INET):
        if isinstance(family, str):
            try:
                family = 'AF_' + family.upper()
                family = getattr(socket, family)
            except AttributeError:
                raise ValueError("Unknown address family '{}'".format(family))
        self.family = family

        if port is None:
            self.address = address
        else:
            warnings.warn(
                    "The 'port' parameter is deprecated, use 'address' in a "
                    "(host, port) format instead.",
                    DeprecationWarning)
            self.address = (address, int(port))

        self.s = None

    def open(self):
        if self.s:
            return
        self.s = socket.socket(self.family, socket.SOCK_STREAM)
        try:
            self.s.connect(self.address)
        except Exception as e:
            raise ConnectorException(e.message)

    def close(self):
        if not self.s:
            return
        self.s.close()
        self.s = None

    def write(self, data):
        try:
            self.s.send(data)
        except socket.timeout as e:
            raise ConnectorException(e.message)

    def read(self, size=1):
        try:
            return self.s.recv(size)
        except socket.timeout:
            # When a recv returns 0 bytes, it means the other side has closed (or is in the process of closing) the connection.
            # Current implementation doesn't differentiate this situation from regular timeout.
            return ""

    def set_read_timeout(self, timeout):
        self.s.settimeout(timeout)

    def set_write_timeout(self, timeout):
        # socket doesn't have separate write timeout
        pass



# Copyright 2021 Silicon Laboratories Inc. www.silabs.com
#
# SPDX-License-Identifier: Zlib
#
# The licensor of this software is Silicon Laboratories Inc.
#
# This software is provided 'as-is', without any express or implied
# warranty. In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

import os
import unittest
from time import sleep

import logging
from testfixtures import log_capture

from bgapi.bglib import BGLib, CommandError, CommandFailedError, BGResponse
from bgapi.apiparser import ParsedApi
from bgapi.serdeser import MSG_COMMAND, MSG_EVENT, HEADER_LENGTH

from .helper import *
from .testconnector import TestConnector

class BGLibTester(unittest.TestCase):

    def setUp(self):
        self.api = ParsedApi(os.path.join(os.path.dirname(__file__), "test.xml"))
        self.conn = TestConnector()
        self.lib = BGLib(self.conn, self.api, log_id="TestLogID")
        self.test = self.lib.test
        self.lib.open()

    def tearDown(self):
        self.lib.close()

    def make_command(self, class_name, command_name, payload):
        cls = self.api[class_name]
        cmd = cls.commands[command_name]
        return make_packet(MSG_COMMAND, self.api.device_id, cls.index, cmd.index, payload)

    def make_event(self, class_name, event_name, payload):
        cls = self.api[class_name]
        evt = cls.events[event_name]
        return make_packet(MSG_EVENT, self.api.device_id, cls.index, evt.index, payload)

    add_step = lambda self, *args, **kwargs: self.conn.add_step(*args, **kwargs)

    def wait(self, time):
        time = float(time)
        self.add_step("@", time)

    def expect_command(self, class_name, command_name, payload):
        self.add_step(">", b"".join(self.make_command(class_name, command_name, payload)))

    def send_response(self, class_name, command_name, payload):
        self.add_step("<", b"".join(self.make_command(class_name, command_name, payload)))

    def send_event(self, class_name, event_name, payload):
        self.add_step("<", b"".join(self.make_event(class_name, event_name, payload)))


class TestCommon(BGLibTester):

    def test_command(self):
        self.expect_command("simple", "returns", "12")
        self.send_response("simple", "returns", "34")

        result = self.lib.test.simple.returns(0x12)
        self.assertEqual(result, (0x34,))

    def test_command_with_success_result(self):
        self.expect_command("simple", "can_fail", "")
        self.send_response("simple", "can_fail", "0000")

        response = self.lib.test.simple.can_fail()
        self.assertEqual(response.result, 0)

    def test_command_no_return(self):
        self.expect_command("simple", "no_return", "12")

        result = self.lib.test.simple.no_return(0x12)
        self.assertEqual(result, None)

    def test_command_complex(self):
        self.expect_command("complex", "multi_type", "125634")
        self.send_response("complex", "multi_type", "abefcd")

        result = self.lib.test.complex.multi_type(0x12, 0x3456)
        self.assertEqual(result, (0xab, 0xcdef))

    def test_command_complex_array(self):
        self.expect_command("complex", "array", "3412025678")
        self.send_response("complex", "array", "cdab02eeff")

        result = self.lib.test.complex.array(0x1234, str2bin("5678"))
        self.assertEqual(result, (0xabcd, str2bin("eeff")))

    def test_event_before_response(self):
        self.expect_command("simple", "returns", "12")
        self.send_event("simple", "no_params", "")
        self.send_response("simple", "returns", "34")

        result = self.lib.test.simple.returns(0x12)
        self.assertEqual(result, (0x34,))
        event = self.lib.get_event()
        self.assertEqual(event.__name__, "NoParams")

    def test_multiple_commands(self):
        self.expect_command("simple", "command_a", "")
        self.send_response("simple", "command_a", "")
        result = self.lib.test.simple.command_a()
        self.assertEqual(result.__name__, "CommandA")

        self.expect_command("simple", "command_b", "")
        self.send_response("simple", "command_b", "")
        result = self.lib.test.simple.command_b()
        self.assertEqual(result.__name__, "CommandB")

    def test_event_generator(self):
        self.send_event("complex", "multi_type_event", "125634")
        self.send_event("simple", "no_params", "")

        eg = self.lib.gen_events(timeout=0.1)

        e = next(eg)
        self.assertEqual(e.__name__, 'MultiTypeEvent')

        e = next(eg)
        self.assertEqual(e.__name__, 'NoParams')

        self.assertRaises(StopIteration, next, eg)

    # Enable debug-level event logs for this test
    @log_capture(level=logging.DEBUG)
    def test_debug_logging(self, l):
        self.expect_command("complex", "multi_type", "010200")
        self.send_response("complex", "multi_type", "020100")
        result = self.lib.test.complex.multi_type(0x01, 0x0002)
        self.send_event("complex", "multi_type_event", "125634")

        event = self.lib.get_event(1.0)

        # Expect log about command, event, and response
        l.check(("bgapi", "DEBUG", "TestLogID > test_cmd_complex_multi_type(param1=1, param2=2)"),
                ("bgapi", "DEBUG", "TestLogID < test_rsp_complex_multi_type(return1=2, return2=1)"),
                ("bgapi", "DEBUG", "TestLogID < test_evt_complex_multi_type_event(value1=18, value2=13398)"))


class TestInvalid(BGLibTester):

    def test_command_no_response(self):
        self.expect_command("simple", "returns", "12")

        self.assertRaises(CommandError, self.lib.test.simple.returns, 0x12)

    def test_command_with_error_result(self):
        self.expect_command("simple", "can_fail", "")
        self.send_response("simple", "can_fail", "3412")

        with self.assertRaisesRegex(CommandFailedError,
                "Command returned 'result' parameter with non-zero errorcode: 0x1234"
                ) as context:
            self.lib.test.simple.can_fail()

        exception = context.exception

        self.assertEqual(exception.errorcode, 0x1234)

        self.assertIsInstance(exception.response, BGResponse)
        self.assertEqual(exception.response.result, 0x1234)

    def test_command_wrong_response(self):
        self.expect_command("simple", "command_a", "")
        self.send_response("simple", "command_b", "")
        self.assertRaises(CommandError, self.lib.test.simple.command_a)

    @log_capture(level=logging.WARNING)
    def test_command_late_response(self, l):
        # Command fails, because no response received in time
        self.expect_command("simple", "command_a", "")
        self.assertRaises(CommandError, self.lib.test.simple.command_a)

        # Send response
        self.send_response("simple", "command_a", "")
        sleep(2 * self.conn.read_timeout) # wait for the response to get processed

        # Expect log about unexpected response
        l.check(("bgapi", "WARNING", "Received unexpected response 'test_rsp_simple_command_a()'"))

        # Send another command and expect to get correct response
        self.expect_command("simple", "command_b", "")
        self.send_response("simple", "command_b", "")
        response = self.lib.test.simple.command_b()
        self.assertEqual(response.__name__, "CommandB")


class TestAPIMismatch(BGLibTester):

    @log_capture(level=logging.WARNING)
    def test_event_with_less_params(self, l):
        self.send_event("complex", "multi_type_event", "12")
        event = self.lib.get_event(1.0)
        self.assertEqual(event.__name__, "MultiTypeEvent")
        self.assertEqual(event, (0x12, None))  # None in place of missing param

        l.check(("bgapi", "WARNING", "Received message 'test_evt_complex_multi_type_event' with parameter(s) 'value2' missing."))

    @log_capture(level=logging.WARNING)
    def test_event_with_more_params(self, l):
        self.send_event("complex", "multi_type_event", "12563478")
        event = self.lib.get_event(1.0)
        self.assertEqual(event.__name__, "MultiTypeEvent")
        self.assertEqual(event, (0x12, 0x3456))  # Extra param is ignored

        l.check(("bgapi", "WARNING", "Received message 'test_evt_complex_multi_type_event' with 1 byte(s) extra payload."))


class TestSplittedRead(BGLibTester):

    def test_splitted_header(self):
        header, payload = self.make_event("simple", "no_params", "")

        self.add_step("<", header[:len(header) // 2]) # send first part of the header
        self.wait(self.conn.read_timeout + 0.1)
        self.add_step("<", header[len(header) // 2:]) # send reset of the header

        event = self.lib.get_event(1.0)
        self.assertEqual(event.__name__, "NoParams")
        self.assertEqual(len(event), 0)

    def test_splitted_packet(self):
        header, payload = self.make_event("simple", "single_uint8", "12")

        self.add_step("<", header) # send header
        self.wait(2 * self.conn.read_timeout)
        self.add_step("<", payload) # send payload

        event = self.lib.get_event(1.0)
        self.assertEqual(event.__name__, "SingleUint8")
        self.assertEqual(event, (0x12,))

    def test_splitted_payload(self):
        header, payload = self.make_event("simple", "double_uint8", "1234")

        self.add_step("<", header + payload[:len(payload) // 2]) # send first part of the packet
        self.wait(2 * self.conn.read_timeout)
        self.add_step("<", payload[len(payload) // 2:]) # send first part of the packet

        event = self.lib.get_event(2.0)
        self.assertEqual(event.__name__, "DoubleUint8")
        self.assertEqual(event, (0x12, 0x34))


class TestOpenClose(BGLibTester):

    def test_close_while_reading_header(self):
        header, payload = self.make_event("simple", "no_params", "")

        self.add_step("<", header[:len(header) // 2])
        self.lib.close()

    def test_close_while_reading_payload(self):
        header, payload = self.make_event("simple", "double_uint8", "1234")

        self.add_step("<", header + payload[:len(payload) // 2])
        self.lib.close()

class TestKeepDeviceAwake(BGLibTester):

    def setUp(self):
        super(TestKeepDeviceAwake, self).setUp()
        self.keep_awake = None
        self.keep_awake_history = []
        self.lib.set_keep_device_awake_function(self.keep_awake_func)

    def keep_awake_func(self, keep_awake):
        self.keep_awake = keep_awake
        self.keep_awake_history.append(keep_awake)

    def test_device_is_not_kept_awake_after_connector_is_opened(self):
        self.lib.close()
        self.keep_awake = 1
        self.lib.open()

        self.assertEqual(self.keep_awake, 0)

    def test_keep_device_awake_when_sending_command(self):
        self.expect_command("simple", "command_a", "")
        self.send_response("simple", "command_a", "")
        self.lib.test.simple.command_a()

        self.assertEqual(self.keep_awake_history, [1, 0])

    def test_device_is_kept_awake_if_command_has_no_response(self):
        self.expect_command("simple", "no_return", "00")
        self.lib.test.simple.no_return(0)

        self.assertEqual(self.keep_awake, 1)

    def test_device_is_not_kept_awake_if_command_fails_to_respond(self):
        self.expect_command("simple", "command_a", "")
        self.assertRaises(CommandError, self.lib.test.simple.command_a)

        self.assertEqual(self.keep_awake, 0)

class TestWithoutLogId(BGLibTester):

    # This class is here to test that missing the optional parameter log_id doesn't break anything.
    def setUp(self):
        self.api = ParsedApi(os.path.join(os.path.dirname(__file__), "test.xml"))
        self.conn = TestConnector()
        self.lib = BGLib(self.conn, self.api)
        self.test = self.lib.test
        self.lib.open()

    @log_capture(level=logging.DEBUG)
    def test_logging_without_log_id(self, l):
        self.expect_command("complex", "multi_type", "010200")
        self.send_response("complex", "multi_type", "020100")
        result = self.lib.test.complex.multi_type(0x01, 0x0002)
        self.send_event("complex", "multi_type_event", "125634")

        event = self.lib.get_event(1.0)
        log_text = str(l).splitlines()
        # If no log_id parameter is given to BGLib, id() will be used, so it'll be a random integer.
        self.assertEqual(log_text[0], "bgapi DEBUG")
        self.assertRegex(log_text[1], '[0-9]+ > test_cmd_complex_multi_type\(param1=1, param2=2\)')
        self.assertEqual(log_text[2], "bgapi DEBUG")
        self.assertRegex(log_text[3], '[0-9]+ < test_rsp_complex_multi_type\(return1=2, return2=1\)')
        self.assertEqual(log_text[4], "bgapi DEBUG")
        self.assertRegex(log_text[5], '[0-9]+ < test_evt_complex_multi_type_event\(value1=18, value2=13398\)')

if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Silicon Laboratories Inc. www.silabs.com
#
# SPDX-License-Identifier: Zlib
#
# The licensor of this software is Silicon Laboratories Inc.
#
# This software is provided 'as-is', without any express or implied
# warranty. In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

def toInt(s):
    if isinstance(s, str):
        s = str(s).strip()
        if s.startswith("0x"):
            return int(s[2:], 16)
        if s.startswith("0b"):
            return int(s[2:], 2)
    return int(s)


class ApiEnumVal(object):

    def __init__(self, enum, node):
        self.parent = enum
        self.name = node.attrib["name"]
        self.value = toInt(node.attrib["value"])
        self.description = node.findtext("description", default="")


class ApiDefineVal(ApiEnumVal): pass


class ApiEnum(dict):

    def __init__(self, api_class, node):
        self.api_class = api_class
        self.name = node.attrib["name"]
        self.description = node.findtext("description", default="")
        self.values = {}
        self.names = []

        for n in node.iter(tag="enum"):
            e = ApiEnumVal(self, n)
            self[e.name] = e
            self[e.value] = e
            self.names.append(e.name)


class ApiDefine(dict):

    def __init__(self, api_class, node):
        self.api_class = api_class
        self.name = node.attrib["name"]
        self.description = node.findtext("description", default="")
        self.values = {}
        self.names = []

        for n in node.iter(tag="enum"):
            e = ApiDefineVal(self, n)
            self.names.append(e.name)
            self[e.name] = e
            self[e.value] = e


class ApiParameter(object):

    def __init__(self, parent, index, node):
        self.parent = parent
        self.index = toInt(index)
        self.name = node.attrib["name"]
        self.format = node.attrib["type"]
        self.datatype = node.get("datatype", None)
        self.validator_type = node.get("validator_type", None)
        self.validator_id = node.get("validator_id", None)
        self.description = node.findtext("description", default="")


class ApiCommand(object):

    def __init__(self, api_class, node):
        self.api_class = api_class
        self.index = toInt(node.attrib["index"])
        self.name = node.attrib["name"]
        self.description = node.findtext("description", default="")
        self.params = []
        self.returns = []
        self.no_return = bool(node.get("no_return", False))
        self.internal = bool(node.get("internal", False))
        if node.find("params") is not None:
            for n in node.find("params").iter(tag="param"):
                self.params.append(ApiParameter(self, len(self.params), n))
        if node.find("returns") is not None:
            for n in node.find("returns").iter(tag="param"):
                self.returns.append(ApiParameter(self, len(self.returns), n))


class ApiEvent(object):

    def __init__(self, api_class, node):
        self.api_class = api_class
        self.index = toInt(node.attrib["index"])
        self.name = node.attrib["name"]
        self.description = node.findtext("description", default="")
        self.internal = bool(node.get("internal", False))
        self.params = []
        for n in node.find("params").iter(tag="param"):
            self.params.append(ApiParameter(self, len(self.params), n))


class ApiClass(object):

    def __init__(self, api, node):
        self.api = api
        self.index = toInt(node.attrib["index"])
        self.name = node.attrib["name"]
        self.description = node.findtext("description", default="")
        self.enums = {}
        self.enum_names = []
        self.defines = {}
        self.define_names = []
        self.commands = {}
        self.command_names = []
        self.events = {}
        self.event_names = []

        for n in node.iter(tag="enums"):
            e = ApiEnum(self, n)
            self.enum_names.append(e.name)
            self.enums[e.name] = e

        for n in node.iter(tag="define"):
            e = ApiDefine(self, n)
            self.define_names.append(e.name)
            self.defines[e.name] = e

        for n in node.iter(tag="command"):
            e = ApiCommand(self, n)
            self.command_names.append(e.name)
            self.commands[e.name] = e
            self.commands[e.index] = e

        for n in node.iter(tag="event"):
            e = ApiEvent(self, n)
            self.event_names.append(e.name)
            self.events[e.name] = e
            self.events[e.index] = e


class ParsedApi(dict):

    def __init__(self, filename=None, deviceId=0):
        self.filename = filename

        self.description = ""
        self.names = []

        from xml.etree.ElementTree import parse
        node = parse(self.filename).getroot()

        self.device_id = toInt(node.attrib["device_id"])
        self.device_name = node.attrib["device_name"]
        self.name = self.device_name
        for n in node.iter(tag="class"):
            e = ApiClass(self, n)
            self.names.append(e.name)
            self[e.name] = e
            self[e.index] = e



# Copyright 2021 Silicon Laboratories Inc. www.silabs.com
#
# SPDX-License-Identifier: Zlib
#
# The licensor of this software is Silicon Laboratories Inc.
#
# This software is provided 'as-is', without any express or implied
# warranty. In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

from array import array
import threading
try:
    import queue
except ImportError:
    import Queue as queue
import time
from time import sleep
import logging

from .apihelper import api_cmd_to_ascii, camelcase
from .apiparser import ParsedApi
from . import serdeser

from .connector import ConnectorTimeoutException


ERRORCODE_SUCCESS = 0

logger = logging.getLogger(__package__)


class CommandError(Exception):
    pass


class CommandFailedError(CommandError):
    """Command response received with a non-zero errorcode return value

    The errorcode received from the device is stored in _errorcode_ attribute.
    The actual response received is stored in _response_ attribute as a
    BGResponse object.
    """

    def __init__(self, response):
        msg = "Command returned '{}' parameter with non-zero errorcode: {:#x}".format(
                response._errorcode_field, response._errorcode)

        super(CommandFailedError, self).__init__(msg)
        self.response = response
        self.errorcode = response._errorcode


class BGApiConnHandler(threading.Thread):
    """This class handles threading when reading and writing to a connection.
    Any responses will go to the response queue, and any events will get
    dispatched to the event handler.
    """

    READ_TIMEOUT = 0.1
    WRITE_TIMEOUT = 1.0

    def __init__(self, connection, event_handler, *apis):
        threading.Thread.__init__(self)
        self.daemon = True # Daemon thread
        self.stop_flag = threading.Event()

        self.conn = connection
        self.event_handler = event_handler

        self.ser = serdeser.Serializer(apis)
        self.deser = serdeser.Deserializer(apis)

        self.response_queue = queue.Queue(maxsize=1)
        self.waiting_response = threading.Event()

    def send_command(self, bgcmd, response_timeout):
        try:
            # Send command
            if not bgcmd.no_response:
                self.waiting_response.set()
            try:
                self.conn.write(self.ser.command(bgcmd._device_name, bgcmd._class_name, bgcmd._msg_name, bgcmd))
            except ConnectorTimeoutException:
                raise CommandError("Send timeout")

            if not bgcmd.no_response:
                # Wait for response
                try:
                    api_node, headers, vals = self.response_queue.get(timeout=response_timeout)
                    response = BGResponse(api_node, vals)

                    if response._errorcode != ERRORCODE_SUCCESS:
                        raise CommandFailedError(response=response)

                    return response
                except queue.Empty:
                    raise CommandError("No response")
        finally:
            self.waiting_response.clear()

    def read(self, size=1):
        data = bytearray()
        while len(data) < size:
            if self.stop_flag.is_set():
                return None
            data.extend(self.conn.read(size - len(data)))
        return bytes(data)

    def run(self):
        self.conn.set_read_timeout(self.READ_TIMEOUT)
        self.conn.set_write_timeout(self.WRITE_TIMEOUT)

        while not self.stop_flag.is_set():
            # This will block
            header = self.read(size=serdeser.HEADER_LENGTH)
            if header is None:
                # Stop flag set while reading header
                continue

            cmdevt, device_id, payload_len, class_index, command_index = self.deser.parseHeader(header)

            # This is the bit that will block some more
            payload = self.read(size=payload_len)
            if payload is None:
                # Stop flag set while reading payload
                continue

            (apicmdevt, headers, params) = self.deser.parse(header, payload, fromHost=False)

            if cmdevt == serdeser.MSG_COMMAND:
                if self.waiting_response.is_set():
                    self.response_queue.put((apicmdevt, headers, params))
                else:
                    # Received response, but not waiting for one. This may
                    # happen if we receive a response to our earlier command
                    # too late.
                    response = BGResponse(apicmdevt, params)
                    logger.warning("Received unexpected response '%s'", response)
            else:
                # Got event
                self.event_handler(BGEvent(apicmdevt, params))

    def stop(self):
        self.stop_flag.set()
        try:
            self.join()
        except RuntimeError:
            pass # ignore error if trying to stop non-started thread


class BGMsg(tuple):
    """Parent class for all command and event classes"""

    def __new__(cls, apinode, values, is_response=False):
        inst = super(BGMsg, cls).__new__(cls, values)
        inst._apinode = apinode
        inst._device_name = apinode.api_class.api.device_name
        inst._class_name = apinode.api_class.name
        inst._msg_name = apinode.name
        inst.__name__ = camelcase(apinode.name)
        inst.__doc__ = api_cmd_to_ascii(apinode)
        inst._timestamp = time.time()

        inst._fields = []
        if hasattr(apinode, "returns") and is_response:
            api_vals = apinode.returns
        else:
            api_vals = apinode.params
        for i in range(len(values)):
            inst._fields.append(api_vals[i].name)
            setattr(inst, api_vals[i].name, values[i])

        return inst

    @property
    def _device_prefix(self):
        return '{}_'.format(self._apinode.api_class.api.device_name)

    @property
    def _type_prefix(self):
        if hasattr(self, 'MSG_TYPE_STRING'):
            return '{}_'.format(self.MSG_TYPE_STRING)
        else:
            return ''

    @property
    def _str(self):
        return "{}{}{}_{}".format(
                self._device_prefix, self._type_prefix,
                self._apinode.api_class.name, self._apinode.name)

    def __repr__(self):
        return "%s(%s)" % (self._str, ", ".join(["%s=%r" % (a, b) for a, b in zip(self._fields, self)]))

    def __eq__(self, other):
        if isinstance(other, str):
            if other.startswith(self._type_prefix):
                # Backwards compatibility if device prefix not provided
                other = self._device_prefix + other
            return self._str == other
        else:
            return super(BGMsg, self).__eq__(other)


class BGEvent(BGMsg):

    MSG_TYPE_STRING = "evt"


class BGCommand(BGMsg):

    MSG_TYPE_STRING = "cmd"

    def __new__(cls, apinode, values):
        inst = super(BGCommand, cls).__new__(cls, apinode, values, is_response=False)
        if apinode.no_return:
            setattr(inst, "no_response", True)
        else:
            setattr(inst, "no_response", False)
        return inst


class BGResponse(BGMsg):

    MSG_TYPE_STRING = "rsp"
    def __new__(cls, apinode, values):
        inst = super(BGResponse, cls).__new__(cls, apinode, values, is_response=True)
        return inst

    @property
    def _errorcode_field(self):
        errorcode_field = None

        for return_field in self._apinode.returns:
            if return_field.datatype == "errorcode":
                errorcode_field = return_field.name
                break

        return errorcode_field

    @property
    def _errorcode(self):
        errorcode = 0

        errorcode_field = self._errorcode_field
        if errorcode_field:
            errorcode = getattr(self, errorcode_field)

        return errorcode


class BGClass(object):

    def __init__(self, name):
        self.__name__ = name

    def __repr__(self):
        return "%s object" % self.__name__.capitalize()


def parse_api(api):
    if isinstance(api, ParsedApi):
        return api
    else:
        return ParsedApi(api)


class BGLib(object):
    """Provides an interface to the Bluegiga binary APIs"""

    def _handle_low_level_event(self, evt):
        logger.debug("{} < {}".format(self.log_id, evt))
        if self.event_handler:
            self.event_handler(evt)
        if not self.event_handler:
            self.event_queue.put(evt)


    def __init__(self, connection, apis, event_handler=None, response_timeout=1, log_id=None):
        """Populates the BGLib with the provided APIs

        The commands are grouped by the API name, defined in the API file. For
        example, if the API is the Silicon Labs Bluetooth API, it will appear
        as an attribute `bt` in the BGLib object:
        >>> bglib.bt.system.hello()

        help() is available for further details, e.g., help(bglib.bt.system)

        :param connection: Instance of a Connector subclass, e.g., bgapi.SerialConnector
        :param apis: A list of API file names or ParsedApi objects
        :param event_handler: A function which takes BGEvent objects, or None in
            which case events are placed in the event queue
        """
        if not isinstance(apis, (list, tuple)):
            apis = [apis]
        apis = [parse_api(api) for api in apis]
        self.log_id = log_id or id(self)

        self.conn = connection
        self.event_handler = event_handler
        self.apis = apis
        self.response_timeout = response_timeout

        self.conn.close() # initially close the connection
        self.conn_handler = None
        self.event_queue = queue.Queue()

        self.set_keep_device_awake_function(None)

        for api in apis:
            apidict = {}
            for class_name in api.names:
                classdict = {}
                #c = makeClass(class_name, api[class_name].description)
                for enum_group_name in api[class_name].enum_names:
                    for enum_name in api[class_name].enums[enum_group_name].names:
                        enum_val = api[class_name].enums[enum_group_name][enum_name]
                        classdict[enum_val.name.upper()] = enum_val.value

                for def_group_name in api[class_name].define_names:
                    for def_name in api[class_name].defines[def_group_name].names:
                        def_val = api[class_name].defines[def_group_name][def_name]
                        classdict[def_val.name.upper()] = def_val.value

                for command_name in api[class_name].command_names:
                    cmd = self._createApiCommand(self, api.device_name, class_name,
                        command_name, api[class_name].commands[command_name])
                    classdict[command_name] = cmd
                classdict["__doc__"] = api[class_name].description
                api_class = type(class_name.capitalize(), (BGClass,), classdict)(class_name)
                apidict[class_name] = api_class

            apidict["__doc__"] = api.description
            setattr(self, api.name, type("%sApi" % api.name.capitalize(), (object,), apidict)())

    def _createApiCommand(self, lib, device_name, class_name, command_name, apicmd):
        response_timeout = self.response_timeout

        def apiCommand(self, *args):
            if not lib.conn_handler:
                raise Exception("Connection is closed")

            lib._keep_device_awake(1)
            try:
                cmd = BGCommand(apicmd, args)
                logger.debug("{} > {}".format(lib.log_id, cmd))
                response = lib.conn_handler.send_command(cmd, response_timeout=response_timeout)
                logger.debug("{} < {}".format(lib.log_id, response))
            except CommandError as e:
                logger.debug("{} ! {}".format(lib.log_id, e))
                raise
            finally:
                if not apicmd.no_return:
                    lib._keep_device_awake(0)

            if apicmd.no_return: return

            # Verify correct response
            if ((response._device_name, response._class_name, response._msg_name) !=
                (cmd._device_name, cmd._class_name, cmd._msg_name)):
                raise CommandError("Wrong response")

            return response

        setattr(apiCommand, "__name__", camelcase(command_name))
        setattr(apiCommand, "_str", "{}_cmd_{}_{}".format(device_name, class_name, command_name))
        setattr(apiCommand, "__doc__", api_cmd_to_ascii(apicmd))

        return apiCommand

    def _keep_device_awake(self, keep_awake):
        if self._keep_device_awake_function:
            self._keep_device_awake_function(keep_awake)

    def set_keep_device_awake_function(self, function):
        """Set a function for keeping device awake

        The function should take one argument, which denotes if the device
        should be kept awake or not. If the value of the argument is zero,
        then the device doesn't need to be kept awake. Any other value means
        that the device should be kept awake.
        """
        self._keep_device_awake_function = function

    def is_open(self):
        return self.conn_handler is not None

    def open(self):
        if not self.is_open():
            self.conn.open()
            self._keep_device_awake(0)
            self.conn_handler = BGApiConnHandler(self.conn, self._handle_low_level_event, *self.apis)
            self.conn_handler.start()

    def close(self):
        if self.is_open():
            self.conn_handler.stop()
            self.conn.close()
            self.conn_handler = None

    def get_event(self, timeout=0.0):
        """Get single event from the device.

        :param timeout: Timeout in seconds to wait for the event, or None for infinite
        :return: BGEvent object, or None if no event received
        """
        try:
            return self.event_queue.get(timeout=timeout)
        except queue.Empty:
            return None

    def get_events(self, timeout=0.0, max_events=None, max_time=None):
        """Get events from the device.

        :param timeout: Timeout in seconds to wait for single event, or None for infinite
        :param max_events: Maximum number of events to receive, or None for no limit
        :param max_time: Maximum time receiving events, or None for no limit
        :return: List of BGEvent objects
        """
        if timeout == None and max_events == None and max_time == None:
            raise Exception("One of the arguments has to be other than None")

        return list(self.gen_events(timeout=timeout, max_events=max_events, max_time=max_time))

    def gen_events(self, timeout=0.0, max_events=None, max_time=None):
        """Generator retrieving events from the device.

        :param timeout: Timeout in seconds to wait for single event, or None for infinite
        :param max_events: Maximum number of events to receive, or None for no limit
        :param max_time: Maximum time receiving events, or None for no limit
        :return: Generator which generates BGEvent objects
        """
        events = 0
        while (max_events == None or events < max_events) and \
              (max_time == None or max_time > 0):

            if max_time != None:
                start_time = time.time()
                if timeout == None or timeout > max_time:
                    timeout = max_time

            event = self.get_event(timeout)
            if event is None:
                break
            yield event
            events += 1

            if max_time:
                max_time -= time.time() - start_time



# Copyright 2021 Silicon Laboratories Inc. www.silabs.com
#
# SPDX-License-Identifier: Zlib
#
# The licensor of this software is Silicon Laboratories Inc.
#
# This software is provided 'as-is', without any express or implied
# warranty. In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

from setuptools import setup
from os import path

with open(path.join(path.dirname(__file__), 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="pybgapi",
    author="Silicon Labs",
    url="https://www.silabs.com",
    description="Python interface for the BGAPI binary protocol",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="zlib",
    classifiers=[
        'License :: OSI Approved :: zlib/libpng License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],

    use_scm_version={'version_scheme': 'post-release'},
    setup_requires=['setuptools_scm'],

    packages=["bgapi"],

    python_requires=">=3.6",
    install_requires=["pyserial"],
    tests_require=["testfixtures"],

    test_suite="bgapi.test",
)
